"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stronger_typed_streams_1 = require("stronger-typed-streams");
const lodash_1 = require("lodash");
const delay = require("delay");
const logger_1 = require("./lib/logger");
const logger = logger_1.default.getLogger('GenericStream');
class GenericStream extends stronger_typed_streams_1.Readable {
    constructor(opt, sdk, method, pathToItems, pathToPageInfo, paginationProp = 'max_id', query, customSuccessHandler) {
        super(opt);
        this.sdk = sdk;
        this.method = method;
        this.pathToItems = pathToItems;
        this.pathToPageInfo = pathToPageInfo;
        this.paginationProp = paginationProp;
        this.query = query;
        this.customSuccessHandler = customSuccessHandler;
        this.requestInProgress = false;
        this.delayMinInterval = 1000;
        this.delayMaxInterval = 3000;
        this.externalBuffer = [];
        this.boundErrorHandler = this.errorHandler.bind(this);
        this.boundSuccessHandler = this.successHandler.bind(this);
        if (this.customSuccessHandler) {
            this.customSuccessHandler = this.customSuccessHandler.bind(this);
        }
    }
    _read(size) {
        while (this.externalBuffer.length && !this.ended && !this.destroyed) {
            const chunk = this.externalBuffer.shift();
            if (chunk && !this.push(chunk)) {
                return;
            }
        }
        if (this.requestInProgress) {
            return;
        }
        this.requestData();
    }
    requestData() {
        this.requestInProgress = true;
        this.getItems(this.endCursor)
            .then(delay(lodash_1.random(this.delayMinInterval, this.delayMaxInterval)))
            .then(this.customSuccessHandler || this.boundSuccessHandler)
            .catch(this.boundErrorHandler);
    }
    getItems(endCursor) {
        const query = Object.assign({}, this.query, { [this.paginationProp]: endCursor });
        return this.sdk[this.method](query);
    }
    destroy(...args) {
        this.externalBuffer = [];
        super.destroy(...args);
    }
    errorHandler(err) {
        this.requestInProgress = false;
        logger.debug('error on fetch %s %s %j', this.method, this.endCursor, this.query, err);
        process.nextTick(() => this.emit('error', err));
    }
    successHandler(response) {
        this.requestInProgress = false;
        const pageInfo = lodash_1.get(response, this.pathToPageInfo, {}), items = lodash_1.get(response, this.pathToItems, []);
        this.endCursor = pageInfo.end_cursor;
        if (!items.length) {
            // sometimes API responses with empty items, but with next page available.
            // re-send request in such case.
            if (pageInfo.has_next_page) {
                delay(lodash_1.random(this.delayMinInterval, this.delayMaxInterval)).then(() => {
                    // FIXME: `this.requestInProgress` - race condition?
                    this.requestData();
                });
                return;
            }
            this.push(null);
            return;
        }
        while (items.length && !this.ended && !this.destroyed) {
            const chunk = items.shift();
            if (chunk && !this.push(chunk)) {
                [].push.apply(this.externalBuffer, items);
                break;
            }
        }
    }
}
exports.GenericStream = GenericStream;
