"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const crypto = require("crypto");
const stream_1 = require("stream");
const util = require("util");
const _ = require("lodash");
const tough = require("tough-cookie");
const RedisCookieStore = require("redis-cookie-store");
const Constants_1 = require("./Constants");
const Signatures_1 = require("./Signatures");
const UserSettings_1 = require("./UserSettings");
const Errors_1 = require("./Errors");
const Request_1 = require("./Request");
const GoodUserAgents_1 = require("./GoodUserAgents");
const logger_1 = require("./lib/logger");
const CookieJar = tough.CookieJar;
class InstagramSDK {
    constructor(username, password, userSettingsDirPath, options = {}) {
        this.username = username;
        this.password = password;
        this.getSelfFollowers = this.getSelfFollowedBy;
        this.getSelfAllFollowers = this.getSelfAllFollowedBy;
        this.appRefreshInterval = 1000 * 60 * 30;
        this.account_id = 0;
        if (!username || !password || !userSettingsDirPath) {
            throw new Error('You must specify both `username`, `password` and `userSettingsDirPath`.');
        }
        this.jar = new CookieJar(new RedisCookieStore(options.redisClient, options.redisPrefix || 'instagram-sdk'));
        this.logLevel = options.LOG_LEVEL || 'INFO';
        this.logger = logger_1.default.getLogger('instagram-sdk', { level: this.logLevel });
        this.settings = new UserSettings_1.default(userSettingsDirPath, username, options.redisClient);
        if (_.isObject(options.proxy) && !_.isEmpty(options.proxy.ip) && !_.isEmpty(options.proxy.port)) {
            this.proxy = options.proxy;
        }
    }
    setup() {
        return __awaiter(this, void 0, void 0, function* () {
            let userAgent = yield this.settings.get('userAgent');
            if (!userAgent) {
                userAgent = GoodUserAgents_1.default.getRandomGoodUserAgent();
                yield this.settings.set('userAgent', userAgent);
            }
            this.userAgent = userAgent;
            let resetCookies = false;
            if (!(yield this.settings.get('uuid')) ||
                !(yield this.settings.get('phone_id')) ||
                !(yield this.settings.get('device_id'))) {
                yield this.settings.set('uuid', Signatures_1.default.generateUUID());
                yield this.settings.set('phone_id', Signatures_1.default.generateUUID());
                yield this.settings.set('device_id', Signatures_1.default.generateDeviceId());
                yield this.settings.set('advertising_id', '');
                yield this.settings.set('account_id', 0);
                yield this.settings.set('token', '');
                yield this.settings.set('last_login', '0');
                resetCookies = true;
            }
            if (!(yield this.settings.get('advertising_id'))) {
                yield this.settings.set('advertising_id', Signatures_1.default.generateUUID());
            }
            this.uuid = yield this.settings.get('uuid');
            this.advertising_id = yield this.settings.get('advertising_id');
            this.device_id = yield this.settings.get('device_id');
            if (!resetCookies && (yield this.settings.isMaybeLoggedIn())) {
                this.isLoggedIn = true;
                this.account_id = yield this.settings.get('account_id');
                this.rank_token = this.account_id + '_' + this.uuid;
                this.token = yield this.settings.get('token');
            }
            else {
                this.isLoggedIn = false;
                this.account_id = 0;
                this.rank_token = null;
                this.token = null;
            }
            if (resetCookies) {
                yield util.promisify(this.jar.store.removeAllCookies.bind(this.jar.store))();
            }
            setInterval(this._sendLoginFlow.bind(this, false), this.appRefreshInterval);
        });
    }
    //////////////////////////
    //// Helpers
    //////////////////////////
    extractCSRFToken({ res }) {
        return __awaiter(this, void 0, void 0, function* () {
            const getCookies = util.promisify(this.jar.getCookies.bind(this.jar)), cookies = yield getCookies(res.url), CSRFCookie = _.find(cookies, { key: 'csrftoken' });
            if (!CSRFCookie) {
                if (!this.token) {
                    this.token = 'missing';
                }
                return res;
            }
            this.token = CSRFCookie.value;
        });
    }
    //////////////////////////
    //// Auth
    //////////////////////////
    getFacebookOTA() {
        return this._request('/facebook_ota/').setBody({
            fields: Constants_1.default.FACEBOOK_OTA_FIELDS,
            custom_user_id: this.account_id,
            signed_body: Signatures_1.default.generateSignature({}) + '.',
            ig_sig_key_version: Constants_1.default.SIG_KEY_VERSION,
            version_code: Constants_1.default.VERSION_CODE,
            version_name: Constants_1.default.IG_VERSION,
            custom_app_id: Constants_1.default.FACEBOOK_ORCA_APPLICATION_ID,
            custom_device_id: this.uuid,
        });
    }
    getExplore() {
        return this._request('/discover/explore/');
    }
    getMegaphoneLog() {
        return this._request('/megaphone/log/').setMethod('POST').setBody(Signatures_1.default.generateSignature({
            type: 'feed_aysf',
            action: 'seen',
            reason: '',
            _uuid: this.uuid,
            device_id: this.device_id,
            _csrftoken: this.token,
            uuid: crypto.createHash('md5').update(Date.now().toString()).digest('hex'),
        }));
    }
    getVisualInbox() {
        return this._request('/direct_v2/visual_inbox');
    }
    getRecentActivity() {
        return this._request('/news/inbox/').setQuery({ activity_module: 'all' });
    }
    getV2Inbox(cursorId = null) {
        const query = {};
        if (cursorId) {
            query.cursor = cursorId;
        }
        return this._request('/direct_v2/inbox/').setQuery(query);
    }
    getRankedRecipients() {
        return this._request('/direct_v2/ranked_recipients').setQuery({ show_threads: true });
    }
    getTimelineFeed(maxId = null) {
        const query = {
            rank_token: this.rank_token,
            ranked_content: true,
        };
        if (maxId) {
            query.max_id = maxId;
        }
        return this._request('/feed/timeline').setQuery(query);
    }
    getRecentRecipients() {
        return this._request('/direct_share/recent_recipients/');
    }
    getReelsTrayFeed() {
        return this._request('/feed/reels_tray/');
    }
    getAutoCompleteUserList() {
        return this._request('/friendships/autocomplete_user_list/')
            .setQuery({ version: '2' })
            .catch((err) => {
            // it's ok, do nothing
        });
    }
    syncFeatures(preLogin = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (preLogin) {
                return yield this._request('/qe/sync/').setMethod('POST').setBody(Signatures_1.default.generateSignature({
                    id: Signatures_1.default.generateUUID(),
                    experiments: Constants_1.default.LOGIN_EXPERIMENTS,
                })).send();
            }
            else {
                return yield this._request('/qe/sync/').setMethod('POST').setBody(Signatures_1.default.generateSignature({
                    _uuid: this.uuid,
                    _uid: this.account_id,
                    _csrftoken: this.token,
                    id: this.account_id,
                    experiments: Constants_1.default.LOGIN_EXPERIMENTS,
                })).send();
            }
        });
    }
    login(force = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isLoggedIn && !force) {
                yield this._sendLoginFlow(false);
                return this;
            }
            yield this.syncFeatures(true);
            yield this._getSignupChallenge();
            const loginResponse = yield this._request('/accounts/login/')
                .setMethod('POST')
                .setBody(Signatures_1.default.generateSignature({
                phone_id: yield this.settings.get('phone_id'),
                device_id: this.device_id,
                guid: this.uuid,
                adid: this.advertising_id,
                username: this.username,
                password: this.password,
                _csrftoken: this.token,
                login_attempt_count: 0,
            })).send();
            yield this._updateLoginState(loginResponse);
            if (loginResponse.status === 'fail' || (_.isObject(loginResponse.errors) && !_.isEmpty(loginResponse.errors))) {
                throw new Errors_1.default.LoginError(loginResponse.message || loginResponse.errors || loginResponse);
            }
            yield this._sendLoginFlow(true);
            return this;
        });
    }
    checkUsername(username) {
        if (!username) {
            throw new Error('Argument `username` is required.');
        }
        /*
         {
         username: 'imposibrus',
         available: false,
         status: 'ok',
         error: 'The username imposibrus is not available.'
         }
         */
        // { username: 'iknergnekjrn', available: true, status: 'ok' }
        return this._request('/users/check_username/')
            .setMethod('POST')
            .setBody(Signatures_1.default.generateSignature({
            _uuid: this.uuid,
            username,
            _csrftoken: 'missing',
        }));
    }
    //////////////////////////
    //// Users
    //////////////////////////
    getSelf() {
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        return this.getUser(this.account_id);
    }
    // tslint:disable-next-line:variable-name
    getSelfRecentMedia(_options = {}) {
        const options = Object.assign({}, _options, { count: 10, min_timestamp: void 0, max_id: void 0 });
        return this.getUserRecentMedia(this.account_id, options);
    }
    getSelfAllMedia(args, { timeout = 0 } = {}) {
        return new Promise((resolve, reject) => {
            let data = [];
            const request = (maxId, cb) => {
                args[0].max_id = maxId;
                this.getSelfRecentMedia.call(this, ...args).then((resp) => {
                    if (resp.status !== 'ok' || (_.isObject(resp.errors) && !_.isEmpty(resp.errors))) {
                        this.logger.error(resp, `(${resp.requestId})`);
                        return reject('Invalid response: ' + JSON.stringify(resp));
                    }
                    data = data.concat(resp.items || []);
                    if (resp.items && resp.items.length) {
                        setTimeout(() => {
                            request(resp.items.slice(-1)[0].id, cb);
                        }, timeout);
                    }
                    else {
                        cb(data);
                    }
                }).catch((err) => {
                    const res = err.res || {};
                    this.logger.error(err, `(${res.requestId})`);
                    reject(err);
                });
            };
            request(void 0, resolve);
        });
    }
    getUser(userID) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        // test/fixtures/userInfo.json
        return this._request(`/users/${userID}/info/`);
    }
    getReelsMediaFeed(usersIDS) {
        if (!usersIDS || !usersIDS.length) {
            throw new Error('Argument `usersIDS` is required and must be not empty.');
        }
        return this._request('/feed/reels_media/')
            .setBody(Signatures_1.default.generateSignature({
            user_ids: usersIDS.map(String),
        }));
    }
    getTagsFeed(tagName) {
        if (!tagName) {
            throw new Error('Argument `tagName` is required and must be not empty.');
        }
        return this._request(`/feed/tag/${tagName}/`);
    }
    // tslint:disable-next-line:variable-name
    getUserRecentMedia(userID, _options = {}) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        const options = Object.assign({}, _options, {
            count: 10,
            min_timestamp: void 0,
            max_id: void 0,
            rank_token: this.rank_token,
            ranked_content: 'true',
        });
        // test/fixtures/userFeed.json
        return this._request(`/feed/user/${userID}/`)
            .setQuery(options);
    }
    getUserAllMedia(args, options) {
        return this._paginate(this.getUserRecentMedia, args, options);
    }
    usersSearch({ q = '', count = 10 } = {}) {
        // test/fixtures/usersSearch.json
        return this._request('/users/search/')
            .setQuery({
            q,
            count,
            rank_token: this.rank_token,
            ig_sig_key_version: Constants_1.default.IG_SIG_KEY,
            is_typeahead: true,
        });
    }
    //////////////////////////
    //// Relationships
    //////////////////////////
    // followed
    getSelfFollows(query = {}) {
        return this.getUserFollows(this.account_id, query);
    }
    getSelfAllFollows(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 0) {
            args.push({});
        }
        return this._paginate(this.getSelfFollows, args, options);
    }
    getSelfFollowsStream(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 0) {
            args.push({});
        }
        return this._stream(this.getSelfFollows, args, options);
    }
    getUserFollows(userID, query = {}) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        query.rank_token = this.rank_token;
        return this._request(`/friendships/${userID}/following/`).setQuery(query);
    }
    getUserAllFollows(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 1) {
            args.push({});
        }
        return this._paginate(this.getUserFollows, args, options);
    }
    getUserFollowsStream(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 1) {
            args.push({});
        }
        return this._stream(this.getUserFollows, args, options);
    }
    // followers
    getSelfFollowedBy(query = {}) {
        return this.getUserFollowers(this.account_id, query);
    }
    getSelfAllFollowedBy(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 0) {
            args.push({});
        }
        return this._paginate(this.getSelfFollowedBy, args, options);
    }
    getSelfFollowersStream(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 0) {
            args.push({});
        }
        return this._stream(this.getSelfFollowedBy, args, options);
    }
    getUserFollowers(userID, query = {}) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        query.rank_token = this.rank_token;
        return this._request(`/friendships/${userID}/followers/`).setQuery(query);
    }
    getUserAllFollowers(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 1) {
            args.push({});
        }
        return this._paginate(this.getUserFollowers, args, options);
    }
    getUserFollowersStream(args, options) {
        options.paginationProp = 'max_id';
        options.dataProp = 'users';
        if (args.length === 1) {
            args.push({});
        }
        return this._stream(this.getUserFollowers, args, options);
    }
    getUserRelationship(userID) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        // tslint:disable-next-line:max-line-length
        // {"status":"ok","incoming_request":false,"outgoing_request":false,"following":false,"followed_by":false,"blocking":false,"is_private":false}
        return this._request(`/friendships/show/${userID}/`);
    }
    getUsersRelationships(usersIDS) {
        if (!usersIDS || !usersIDS.length) {
            throw new Error('Argument `usersIDS` is required and must be not empty.');
        }
        // test/fixtures/friendshipsMany.json
        return this._request('/friendships/show_many/')
            .setMethod('POST')
            .setBody({
            user_ids: usersIDS.join(','),
            _uuid: this.uuid,
            _csrftoken: this.token,
        });
    }
    followUser(userID) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        return this._request(`/friendships/create/${userID}/`)
            .setMethod('POST')
            .setBody(Signatures_1.default.generateSignature({
            _uuid: this.uuid,
            _uid: this.account_id,
            user_id: userID,
            _csrftoken: this.token,
            radio_type: 'wifi-none',
        }));
    }
    unFollowUser(userID) {
        if (!userID) {
            throw new Error('Argument `userID` is required.');
        }
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        return this._request(`/friendships/destroy/${userID}/`)
            .setMethod('POST')
            .setBody(Signatures_1.default.generateSignature({
            _uuid: this.uuid,
            _uid: this.account_id,
            _csrftoken: this.token,
            user_id: userID,
            radio_type: 'wifi-none',
        }));
    }
    //////////////////////////
    //// Media
    //////////////////////////
    getMediaInfoById(mediaID) {
        if (!mediaID) {
            throw new Error('Argument `mediaID` is required.');
        }
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        return this._request(`/media/${mediaID}/info/`);
    }
    getCommentsForMedia(mediaID, maxId = null) {
        if (!mediaID) {
            throw new Error('Argument `mediaID` is required.');
        }
        if (!this.isLoggedIn) {
            throw new Error('You must be logged in.');
        }
        // test/fixtures/mediaComments.json
        return this._request(`/media/${mediaID}/comments/`)
            .setBody({
            ig_sig_key_version: Constants_1.default.IG_SIG_KEY,
            max_id: maxId,
        });
    }
    getLikesForMedia(mediaID) {
        if (!mediaID) {
            throw new Error('Argument `mediaID` is required.');
        }
        // test/fixtures/likers.json
        return this._request(`/media/${mediaID}/likers/`);
    }
    addLikeForMedia(mediaID) {
        if (!mediaID) {
            throw new Error('Argument `mediaID` is required.');
        }
        // {"status":"ok"}
        return this._request(`/media/${mediaID}/like/`)
            .setMethod('POST')
            .setBody(Signatures_1.default.generateSignature({
            _uuid: this.uuid,
            _uid: this.account_id,
            _csrftoken: this.token,
            media_id: mediaID,
        }));
    }
    removeLikeForMedia(mediaID) {
        if (!mediaID) {
            throw new Error('Argument `mediaID` is required.');
        }
        // {"status":"ok"}
        return this._request(`/media/${mediaID}/unlike/`)
            .setMethod('POST')
            .setBody(Signatures_1.default.generateSignature({
            _uuid: this.uuid,
            _uid: this.account_id,
            _csrftoken: this.token,
            media_id: mediaID,
        }));
    }
    tagsSearch(q = '', count = 20) {
        return this._request('/tags/search/').setQuery({ q, count, rank_token: this.rank_token });
    }
    _request(url = '/users/self') {
        return new Request_1.default(this, url);
    }
    _sendLoginFlow(justLoggedIn) {
        if (justLoggedIn) {
            return this.syncFeatures()
                .then(() => this.getAutoCompleteUserList())
                .then(() => this.getReelsTrayFeed())
                .then(() => this.getRecentRecipients())
                .then(() => this.getTimelineFeed())
                .then(() => this.getRankedRecipients())
                .then(() => this.getV2Inbox())
                .then(() => this.getRecentActivity())
                .then(() => this.getVisualInbox())
                .then(() => this.getExplore());
            // .then(() => this.getFacebookOTA())
        }
        else {
            return this.getTimelineFeed().then((data) => {
                if (data.message === 'login_required') {
                    return this.login(true);
                }
                this.settings.set('last_login', Date.now());
                return this.getAutoCompleteUserList()
                    .then(() => this.getReelsTrayFeed())
                    .then(() => this.getRankedRecipients())
                    .then(() => this.getRecentRecipients())
                    .then(() => this.getMegaphoneLog())
                    .then(() => this.getV2Inbox())
                    .then(() => this.getRecentActivity())
                    .then(() => this.getExplore());
            });
        }
    }
    _getSignupChallenge() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._request('/si/fetch_headers/').setQuery({
                challenge_type: 'signup',
                guid: this.uuid.replace(/-/g, ''),
            }).setParseResponseInJSON(false).send();
        });
    }
    _updateLoginState(data) {
        return __awaiter(this, void 0, void 0, function* () {
            this.isLoggedIn = true;
            this.account_id = data.logged_in_user.pk;
            yield this.settings.set('account_id', this.account_id);
            this.rank_token = this.account_id + '_' + this.uuid;
            yield this.settings.set('token', this.token);
            yield this.settings.set('last_login', Date.now());
            return data;
        });
    }
    _paginate(method = this.getSelfRecentMedia, methodArgs = [], { timeout = 0, paginationProp = 'max_id', dataProp = 'items', } = {}) {
        return new Promise((resolve, reject) => {
            let data = [];
            const request = (maxId, cb) => {
                if (methodArgs.length === 1) {
                    methodArgs[0][paginationProp] = maxId;
                }
                else {
                    const lastArg = methodArgs.slice(-1)[0];
                    lastArg[paginationProp] = maxId;
                    methodArgs.splice(-1, 1, lastArg);
                }
                method.call(this, ...methodArgs).then((resp) => {
                    if (resp.status !== 'ok' || (_.isObject(resp.errors) && !_.isEmpty(resp.errors))) {
                        this.logger.error(resp, `(${resp.requestId})`);
                        return reject('Invalid response: ' + JSON.stringify(resp));
                    }
                    data = data.concat(resp[dataProp]);
                    if (resp.pagination && resp.pagination['next_' + paginationProp]) {
                        setTimeout(() => {
                            request(resp.pagination['next_' + paginationProp], cb);
                        }, timeout);
                    }
                    else if (resp['next_' + paginationProp]) {
                        setTimeout(() => {
                            request(resp['next_' + paginationProp], cb);
                        }, timeout);
                    }
                    else {
                        cb(data);
                    }
                }).catch((err) => {
                    const res = err.res || {};
                    this.logger.error(err, `(${res.requestId})`);
                    reject(err);
                });
            };
            request(void 0, resolve);
        });
    }
    // FIXME: rewrite to normal streams
    _stream(method = this.getSelfRecentMedia, methodArgs = [], { timeout = 0, paginationProp = 'max_id', dataProp = 'items', } = {}) {
        const rs = new stream_1.Readable({
            objectMode: true,
            read: () => {
                if (rs._closing) {
                    return false;
                }
            },
        });
        rs._closing = false;
        rs.close = () => {
            rs._closing = true;
            process.nextTick(() => {
                rs.push(null);
                // hacking private API :(
                rs._readableState.ended = true;
                rs._readableState.length = 0;
                rs._readableState.buffer = [];
            });
        };
        const request = (maxId) => {
            if (methodArgs.length === 1) {
                methodArgs[0][paginationProp] = maxId;
            }
            else {
                const lastArg = methodArgs.slice(-1)[0];
                lastArg[paginationProp] = maxId;
                methodArgs.splice(-1, 1, lastArg);
            }
            method.call(this, ...methodArgs).then((resp) => {
                const items = resp[dataProp];
                if (resp.status !== 'ok' || !_.isArray(items)) {
                    this.logger.error(resp, `(${resp.requestId})`);
                    return rs.emit('error', new Error('Invalid response: ' + JSON.stringify(resp)));
                }
                for (const item of items) {
                    if (rs._closing) {
                        break;
                    }
                    rs.push(item);
                }
                if (rs._closing) {
                    return false;
                }
                if (resp.pagination && resp.pagination['next_' + paginationProp]) {
                    setTimeout(() => {
                        request(resp.pagination['next_' + paginationProp]);
                    }, timeout);
                }
                else if (resp['next_' + paginationProp]) {
                    setTimeout(() => {
                        request(resp['next_' + paginationProp]);
                    }, timeout);
                }
                else {
                    if (!rs._closing) {
                        process.nextTick(() => {
                            rs.push(null);
                        });
                    }
                }
            }).catch((err) => {
                const res = err.res || {};
                this.logger.error(err);
                rs.emit('error', err, `(${res.requestId})`);
            });
        };
        request(void 0);
        return rs;
    }
}
InstagramSDK.Errors = Errors_1.default;
exports.InstagramSDK = InstagramSDK;
