"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const util = require("util");
const _ = require("lodash");
const tough = require("tough-cookie");
const BigNumber = require("big-number");
const RequestWeb_1 = require("./RequestWeb");
const Errors_1 = require("./Errors");
const GenericStream_1 = require("./GenericStream");
class IGSDK {
    constructor(options = {}) {
        this.options = {};
        this.logLevel = process.env.LOG_LEVEL || 'INFO';
        this.defaultOptions = {};
        this.options = Object.assign({}, this.defaultOptions, options);
        this.jar = new tough.CookieJar(options.cookieJarStore);
        this.getCookies = util.promisify(this.jar.getCookies.bind(this.jar));
        this.getCookieString = util.promisify(this.jar.getCookieString.bind(this.jar));
        this.setCookie = util.promisify(this.jar.setCookie.bind(this.jar));
    }
    static idToShortCode(id) {
        let photoId = new BigNumber(id.toString().split('_')[0]), code = '';
        while (photoId.gt(0)) {
            const prevId = photoId.val(), remainder = Number(photoId.mod(64));
            photoId = new BigNumber(prevId).minus(remainder).div(64);
            code = this.alphabet[remainder] + code;
        }
        return code;
    }
    static shortCodeToId(code) {
        let id = new BigNumber(0);
        for (const char of code) {
            id = id.multiply(64).plus(this.alphabet.indexOf(char));
        }
        return id.toString();
    }
    static withCredentials(username, password, options) {
        const instance = new IGSDK(options);
        instance.username = username;
        instance.password = password;
        return instance;
    }
    static genHeaders(token, cookies) {
        const CSRFToken = token || _.get(cookies.find((cookie) => cookie.key === 'csrftoken'), 'value');
        return {
            referer: `${RequestWeb_1.Request.baseUrl}/`,
            'x-csrftoken': CSRFToken,
        };
    }
    extractCSRFToken(res) {
        return __awaiter(this, void 0, void 0, function* () {
            const cookies = yield this.getCookies(res.url), CSRFCookie = cookies.find((cookie) => cookie.key === 'csrftoken');
            if (!CSRFCookie) {
                if (!this.token) {
                    this.token = 'missing';
                }
                return;
            }
            this.token = CSRFCookie.value;
        });
    }
    makeStream(method, pathToItems, pathToPageInfo, paginationProp, query, customSuccessHandler) {
        return new GenericStream_1.GenericStream({ objectMode: true, highWaterMark: 1 }, this, method, pathToItems, pathToPageInfo, paginationProp, query, customSuccessHandler);
    }
    getUserByName(userName) {
        return this.request(`/${userName}/`)
            .setQuery({ __a: 1 })
            .send();
    }
    login(force = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!force && (yield this.isLoggedIn())) {
                return true;
            }
            const req = this.request(`/`)
                .setMethod('HEAD')
                .setParseResponseInJSON(false);
            yield req.send();
            if (!req.response) {
                throw new Error('request has been failed');
            }
            const cookies = RequestWeb_1.Request.parseCookies(req.response.headers);
            return this.request(`/accounts/login/ajax/`)
                .setMethod('POST')
                .setHeaders(IGSDK.genHeaders(this.token, cookies))
                .setBody({
                username: this.username,
                password: this.password,
            })
                .send();
        });
    }
    isLoggedIn() {
        return __awaiter(this, void 0, void 0, function* () {
            const cookies = yield this.getCookies(`${RequestWeb_1.Request.baseUrl}/`), sessionIdCookie = cookies.find((cookie) => cookie.key === 'sessionid');
            if (!sessionIdCookie) {
                return false;
            }
            const req = this.request(`/`)
                .setMethod('HEAD')
                .setHeaders(IGSDK.genHeaders(this.token, cookies))
                .setParseResponseInJSON(false);
            yield req.send();
            if (!req.response) {
                return false;
            }
            const responseCookies = RequestWeb_1.Request.parseCookies(req.response.headers), userIdCookie = responseCookies.find((cookie) => cookie.key === 'ds_user_id');
            return !!userIdCookie;
        });
    }
    getPostById(id) {
        const code = IGSDK.idToShortCode(id);
        return this.getPostByShortCode(code);
    }
    getPostByShortCode(shortCode) {
        return this.request(`/p/${shortCode}/`)
            .setQuery({ __a: 1 })
            .send();
    }
    getLocationById(locationId) {
        return this.request(`/explore/locations/${locationId}/`)
            .setQuery({ __a: 1, max_id: '' })
            .send();
    }
    graphQLQuery(opts, queryId) {
        return this.request('/graphql/query/')
            .setQuery({
            query_id: queryId,
            variables: JSON.stringify(opts),
        })
            .send();
    }
    getUserFollowers(opts) {
        return this.graphQLQuery(opts, '17851374694183129');
    }
    getUserFollows(opts) {
        return this.graphQLQuery(opts, '17874545323001329');
    }
    getUserPosts(opts) {
        return this.graphQLQuery(opts, '17888483320059182');
    }
    makeUserFollowersSteam(opts) {
        return this.makeStream('getUserFollowers', 'data.user.edge_followed_by.edges', 'data.user.edge_followed_by.page_info', 'after', opts);
    }
    makeUserFollowsSteam(opts) {
        return this.makeStream('getUserFollows', 'data.user.edge_follow.edges', 'data.user.edge_follow.page_info', 'after', opts);
    }
    makeUserPostsSteam(opts) {
        return this.makeStream('getUserPosts', 'data.user.edge_owner_to_timeline_media.edges', 'data.user.edge_owner_to_timeline_media.page_info', 'after', opts);
    }
    request(path) {
        return new RequestWeb_1.Request(this, path);
    }
}
IGSDK.Errors = Errors_1.default;
IGSDK.alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_';
exports.IGSDK = IGSDK;
var GenericStream_2 = require("./GenericStream");
exports.GenericStream = GenericStream_2.GenericStream;
