"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const util = require("util");
const _ = require("lodash");
const tough = require("tough-cookie");
const got = require("got");
const uuid = require("uuid");
// import * as tunnel from 'tunnel';
const Errors_1 = require("./Errors");
const logger_1 = require("./lib/logger");
const Cookie = tough.Cookie;
class Request {
    constructor(sdk, url) {
        this.sdk = sdk;
        this.requestId = uuid.v4();
        this.method = 'GET';
        this.body = {};
        this.query = {};
        this.headers = {};
        this.defaultRequestOptions = {
            decompress: true,
            followRedirect: true,
        };
        this.checkStatusCode = true;
        this.successStatusCodes = new Set([200, 204]);
        this.parseResponseInJSON = true;
        this.reloginOnError = true;
        this.url = Request.baseUrl + url;
        this.logger = logger_1.default.getLogger('instagram-sdk:Request', { level: this.sdk.logLevel });
    }
    static parseCookies(headers) {
        const setCookieHeader = _.get(headers, 'set-cookie', ''), filteredCookies = [];
        let parsedCookies;
        if (Array.isArray(setCookieHeader)) {
            parsedCookies = setCookieHeader.map((str) => Cookie.parse(str));
        }
        else {
            parsedCookies = [Cookie.parse(setCookieHeader)];
        }
        for (const cookie of parsedCookies) {
            if (!cookie) {
                continue;
            }
            filteredCookies.push(cookie);
        }
        return filteredCookies;
    }
    static _normalizeQueryParams(query) {
        const out = {};
        for (const i in query) {
            if (Object.prototype.hasOwnProperty.call(query, i)) {
                if (query[i] !== undefined && query[i] !== null) {
                    out[i] = query[i];
                }
            }
        }
        return out;
    }
    setReloginOnError(relogin) {
        this.reloginOnError = relogin;
        return this;
    }
    setParseResponseInJSON(parse) {
        this.parseResponseInJSON = parse;
        return this;
    }
    setSuccessStatusCodes(codes) {
        this.successStatusCodes = new Set(codes);
        return this;
    }
    setCheckStatusCode(check) {
        this.checkStatusCode = check;
        return this;
    }
    setMethod(method) {
        this.method = method;
        return this;
    }
    setBody(body) {
        this.body = body;
        return this;
    }
    setQuery(query) {
        this.query = query;
        return this;
    }
    setHeaders(headers) {
        this.headers = headers;
        return this;
    }
    getResultHeaders() {
        return __awaiter(this, void 0, void 0, function* () {
            const cookie = yield this.sdk.getCookieString(this.url);
            let headers = this.headers;
            if (cookie) {
                headers = Object.assign({}, this.headers, { cookie });
            }
            if (_.isEmpty(headers)) {
                return undefined;
            }
            return headers;
        });
    }
    getResultQuery() {
        const query = Request._normalizeQueryParams(this.query);
        if (_.isEmpty(query)) {
            return undefined;
        }
        return query;
    }
    getRequestOptions(headers, query) {
        const opts = {
            method: this.method,
            json: true,
            headers,
            query,
        };
        return Object.assign({}, this.defaultRequestOptions, opts);
    }
    addBodyParam(key, value) {
        this.body[key] = value;
        return this;
    }
    addQueryParam(key, value) {
        this.query[key] = value;
        return this;
    }
    addHeader(key, value) {
        this.headers[key] = value;
        return this;
    }
    send() {
        return __awaiter(this, void 0, void 0, function* () {
            const headers = yield this.getResultHeaders(), query = this.getResultQuery(), options = this.getRequestOptions(headers, query);
            // if (this.sdk.proxy) {
            //     options.agent = tunnel.httpOverHttp({
            //         proxy: {
            //             host: this.sdk.proxy.ip,
            //             port: this.sdk.proxy.port,
            //             headers: {},
            //         },
            //     });
            //     this.logger.debug(`(${this.requestId}) Request.send: using proxy: %j`, this.sdk.proxy);
            // }
            if (!_.isEmpty(this.body)) {
                options.body = this.body;
                options.form = true;
                this.logger.debug(`(${this.requestId}) Request.send: request with body: %:2j`, this.body);
            }
            this.logger.debug(`(${this.requestId}) Request.send: request to %s with params: %:2j`, this.url, _.omit(options, ['body']));
            try {
                const res = yield got(this.url, options);
                this.response = res;
                res.requestId = this.requestId;
                this.logger.debug(`(${this.requestId}) Request.send: response statusCode:`, res.statusCode);
                this.logger.debug(`(${this.requestId}) Request.send: response headers: %:2j`, res.headers);
                const cookies = Request.parseCookies(res.headers);
                for (const cookie of cookies) {
                    yield this.sdk.setCookie(cookie, this.url);
                }
                yield this.sdk.extractCSRFToken(res);
                // if (this.reloginOnError && res.body.message === 'login_required') {
                //     this.logger.debug(
                //         `(${this.requestId}) Request.send: 'login_required', try again: %j`,
                //         res.body,
                //     );
                //
                //     return this.sdk.login(true).then(() => {
                //         return this.send();
                //     });
                // }
                if (this.parseResponseInJSON) {
                    this.logger.debug(`(${this.requestId}) Request.send parsing JSON: response: %.-500s`, util.inspect(res.body, { depth: 5 }));
                }
                if (res.body) {
                    res.body.requestId = this.requestId;
                }
                return Promise.resolve(res.body);
            }
            catch (err) {
                if (err.response) {
                    err.response.requestId = this.requestId;
                }
                err.res = err.response;
                if (err instanceof got.HTTPError) {
                    if (this.checkStatusCode && !this.successStatusCodes.has(err.statusCode)) {
                        if (err.statusCode === 429) {
                            if (this.sdk.options.repeatOnTooManyRequestsInterval) {
                                return delay(this.send.bind(this), this.sdk.options.repeatOnTooManyRequestsInterval);
                            }
                            const error = new Errors_1.default.ThrottledError();
                            error.res = err.response;
                            throw error;
                        }
                    }
                    else {
                        this.logger.debug('request error', err);
                        return Promise.resolve(err.response);
                    }
                }
                throw err;
            }
        });
    }
}
Request.baseUrl = 'https://www.instagram.com';
exports.Request = Request;
function delay(func, timeout) {
    return new Promise((resolve, reject) => {
        setTimeout(() => {
            func().then(resolve).catch(reject);
        }, timeout);
    });
}
